#!/usr/bin/ruby
require 'socket'
require 'thread'
require './common'

$server_host  = ARGV.shift
$server_port  = ARGV.shift
$port         = ARGV.shift
$upload_dir   = "./up/"
$download_dir = "./down/"

class P4UClient
  def initialize
    # OC
    @server = TCPSocket.open($server_host, $server_port)
    @listen = TCPServer.open($port)
    $host = @server.gets.chop
    puts("+ accepted")
    # t@Cꗗɒǉ
    @uploads = []
    update
  end
  
  def main
    t = Thread.new {
      send_listen
    }
    local_session
    t.kill
  end
  
  def finalize
    # t@Cꗗ폜
    while @uploads.size > 0 do
      info = @uploads.pop
      del_file(info)
    end
    # \Pbg
    @listen.close if @listen
    @server.close if @server
    puts("+ closed")
  end
  
  # MZbV
  def send_listen
    while true do
      Thread.new(@listen.accept) {|s|
        send_session(s)
      }
    end
  end
  
  # MZbV
  def send_session(sock)
    begin
      # t@C𒊏o
      digest = sock.gets.chop
      info = @uploads.find {|info|
        info.digest == digest
      }
      # t@C𑗐M
      puts("+ sending #{info.name}")
      File.open($upload_dir + info.name, "rb") {|f|
      	pos = sock.gets.to_i
      	f.seek(pos, IO::SEEK_SET)
      	remain = info.size - pos
        while remain > 0 do
          data = f.read([65536, remain].min)
          sock.write(data)
          remain -= [65536, remain].min
        end
      }
      puts("+ #{info.name} sended")
    rescue
      puts("! #{info.name} sending failed")
    ensure
      sock.close
    end
  end
  
  # MZbV
  def recv_session(digest)
    begin
      # t@Co
      @server.puts("INFO #{digest}")
      size = @server.gets.chop.to_i
      array = []
      size.times {
        info = @server.gets.chop.to_info
        array.push(info)
      }
      info = array.first
      # t@C񑗐M
      puts("+ receiving #{info.name}")
      sock = TCPSocket.open(info.host, info.port)
      sock.puts(info.digest)
      # t@CM
      mode = File::WRONLY | File::CREAT | File::BINARY
      File.open($download_dir + info.name, mode) {|f|
      	f.seek(0, IO::SEEK_END)
      	pos = f.tell
      	sock.puts(pos.to_s)
        remain = info.size - pos
        while remain > 0 do
          data = sock.read([65536, remain].min)
          f.write(data)
          remain -= [65536, remain].min
        end
      }
      sock.close
      puts("+ #{info.name} received")
    rescue
      puts("! #{info.name} receiving failed")
    end
  end
  
  # [JZbV
  def local_session
    while line = gets do
      case line.chop
      when /^update$/
        update
      when /^ls(?:\s+(.*))?$/
        dir = $1 || "."
        get_list(dir)
      when /^lls(?:\s+(.*))?$/
        dir = $1 || "."
        local_list(dir)
      when /^get\s+([0-9a-f]{32})$/
        Thread.new($1) {|digest|
          recv_session(digest)
        }
      when /^quit$/
        break
      else
        puts("! undefined operation")
      end
    end
  end
  
  # t@CXgXV
  def update
    latest = []
    before = @uploads.clone
    # Xgo
    Dir.open($upload_dir).each {|f|
      next if File.directory?(f)
      info = FileInfo.create(f)
      latest.push(info)
    }
    # 폜
    (before - latest).each {|info|
      del_file(info)
      @uploads.delete(info)
    }
    # ǉ
    (latest - before).each {|info|
      add_file(info)
      @uploads.push(info)
    }
  end
  
  # t@C
  def get_list(word)
    @server.puts("LS #{word}")
    size = @server.gets.chop.to_i
    size.times {
      info = @server.gets.chop.to_info
      puts("+ #{info.name} (#{info.digest}) #{info.size} bytes")
    }
    puts("! no such a file") if size == 0
  end
  
  # [J
  def local_list(word)
    array = @uploads.clone
    array.delete_if {|info|
      not info.name.include?(word)
    }
    array.each {|info|
      puts("+ #{info.name} (#{info.digest}) #{info.size} bytes")
    }
    puts("! no such a file") if array.empty?
  end
  
  # t@Cǉ
  def add_file(info)
    if info then
      data = info.to_string
      @server.puts("ADD #{data}")
      puts(@server.gets)
    else
      puts("! no such a file")
    end
  end
  
  # t@C폜
  def del_file(info)
    if info then
      data = info.to_string
      @server.puts("DEL #{data}")
      puts(@server.gets)
    else
      puts("! no such a file")
    end
  end
end

begin
  client = P4UClient.new
  client.main
rescue
  puts("#{$!}")
ensure
  client.finalize if client
  exit(0)
end
